import os
import re
import textwrap

from bs4 import BeautifulSoup, Tag, formatter
from my_app_name._zrb.config import APP_DIR

from zrb.context.any_context import AnyContext
from zrb.util.codemod.modify_class_property import append_property_to_class
from zrb.util.file import read_file, write_file
from zrb.util.string.conversion import (
    to_human_case,
    to_kebab_case,
    to_pascal_case,
    to_snake_case,
)


def update_my_app_name_schema(ctx: AnyContext):
    snake_entity_name = to_snake_case(ctx.input.entity)
    pascal_entity_name = to_pascal_case(ctx.input.entity)
    snake_column_name = to_snake_case(ctx.input.column)
    column_type = ctx.input.type
    schema_file_path = os.path.join(APP_DIR, "schema", f"{snake_entity_name}.py")
    existing_code = read_file(schema_file_path)
    # Base
    new_code = append_property_to_class(
        original_code=existing_code,
        class_name=f"{pascal_entity_name}Base",
        property_name=snake_column_name,
        annotation=column_type,
        default_value=_get_default_column_value(column_type),
    )
    # Update
    new_code = append_property_to_class(
        original_code=new_code,
        class_name=f"{pascal_entity_name}Update",
        property_name=snake_column_name,
        annotation=f"{column_type} | None",
        default_value="None",
    )
    # Table
    new_code = append_property_to_class(
        original_code=new_code,
        class_name=f"{pascal_entity_name}",
        property_name=snake_column_name,
        annotation=f"{column_type} | None",
        default_value="Field(index=False)",
    )
    write_file(schema_file_path, new_code)


def _get_default_column_value(data_type: str) -> str:
    if data_type == "str":
        return '""'
    if data_type in ("int", "float"):
        return "0"
    if data_type == "bool":
        return "True"
    return "None"


def update_my_app_name_ui(ctx: AnyContext):
    kebab_module_name = to_kebab_case(ctx.input.module)
    kebab_entity_name = to_kebab_case(ctx.input.entity)
    snake_column_name = to_snake_case(ctx.input.column)
    human_column_name = to_human_case(ctx.input.column).title()
    column_type = ctx.input.type
    subroute_file_path = os.path.join(
        APP_DIR,
        "module",
        "gateway",
        "view",
        "content",
        kebab_module_name,
        f"{kebab_entity_name}.html",
    )
    existing_code = read_file(subroute_file_path)
    # Add table header
    new_code = _add_th_before_last(
        existing_code, table_id="crud-table", th_content=human_column_name
    )
    # Forms
    new_code = _add_input_to_form(
        new_code,
        form_id="crud-create-form",
        column_label=human_column_name,
        column_name=snake_column_name,
        column_type=column_type,
    )
    new_code = _add_input_to_form(
        new_code,
        form_id="crud-update-form",
        column_label=human_column_name,
        column_name=snake_column_name,
        column_type=column_type,
    )
    new_code = _add_input_to_form(
        new_code,
        form_id="crud-delete-form",
        column_label=human_column_name,
        column_name=snake_column_name,
        column_type=column_type,
    )
    # JS Function
    new_code = _alter_js_function_returned_array(
        new_code,
        js_function_name="getRowComponents",
        js_array_name="rowComponents",
        js_new_value=f"`<td>${{row.{snake_column_name}}}</td>`",
    )
    write_file(subroute_file_path, new_code)


def _add_th_before_last(html_str, table_id, th_content):
    # Use the html.parser; you might try html5lib if you find that it preserves formatting better.
    soup = BeautifulSoup(html_str, "html.parser")
    # Locate the table with the specified id
    table = soup.find("table", id=table_id)
    if not table:
        # Table not found; return original HTML unchanged.
        return html_str
    # Find the thead element in the table.
    thead = table.find("thead")
    if not thead:
        return html_str
    # For this example, we assume there's a single row (<tr>) in the thead.
    row = thead.find("tr")
    if not row:
        return html_str
    # Find all existing th elements in the row.
    th_elements = row.find_all("th")
    if not th_elements:
        return html_str
    # Create a new th element and set its content.
    new_th = soup.new_tag("th")
    new_th.string = th_content
    # Insert the new th right before the last existing th.
    th_elements[-1].insert_before(new_th)
    # Return the modified HTML as a string.
    return soup.prettify(
        formatter=formatter.HTMLFormatter(indent=_infer_html_indent_width(html_str))
    )


def _add_input_to_form(
    html_str: str, form_id: str, column_label: str, column_name: str, column_type: str
) -> str:
    soup = BeautifulSoup(html_str, "html.parser")
    # Find the form by id.
    form = soup.find("form", id=form_id)
    if not form:
        return html_str  # Return unchanged if no matching form is found.
    # Create a new label element with the provided column label.
    new_label = soup.new_tag("label")
    new_label.append(f"{column_label}: ")
    # Create a new input element with the provided column name.
    new_input = _get_html_input(soup, column_name, column_type)
    new_label.append(new_input)
    # Look for a footer element inside the form.
    footer = form.find("footer")
    if footer:
        # Insert the new label before the footer.
        footer.insert_before(new_label)
    else:
        # If no footer exists, simply append the new label to the form.
        form.append(new_label)
    return soup.prettify(
        formatter=formatter.HTMLFormatter(indent=_infer_html_indent_width(html_str))
    )


def _get_html_input(soup: BeautifulSoup, column_name: str, column_type: str) -> Tag:
    # Map your custom types to HTML input types.
    type_mapping = {
        "str": "text",
        "int": "number",
        "float": "number",
        "bool": "checkbox",
        "datetime": "datetime-local",
        "date": "date",
    }
    # Get the corresponding HTML input type; default to "text" if not found.
    html_input_type = type_mapping.get(column_type, "text")
    # Create the new input tag with the appropriate attributes.
    new_input = soup.new_tag(
        "input",
        attrs={"type": html_input_type, "name": column_name, "required": "required"},
    )
    return new_input


def _infer_html_indent_width(html_str: str) -> int:
    """
    Infer the indentation width (number of spaces) from the HTML string.
    It looks for the first non-empty line that starts with whitespace
    followed by '<' and returns the number of leading spaces.
    If none is found, defaults to 2.
    """
    for line in textwrap.dedent(html_str).splitlines():
        stripped = line.lstrip()
        if stripped.startswith("<") and line != stripped:
            return len(line) - len(stripped)
    return 2


def _alter_js_function_returned_array(
    html_str: str, js_function_name: str, js_array_name: str, js_new_value: str
) -> str:
    """
    Inserts a new push into the specified JavaScript function.

    It finds the function definition with the given js_function_name,
    then looks for the first return statement inside the function body,
    and inserts a new line that pushes js_new_value into js_arr_name.

    Parameters:
        html_str (str): The HTML containing the JavaScript.
        js_function_name (str): The name of the JavaScript function to modify.
        js_arr_name (str): The name of the array inside that function.
        js_new_value (str): The new value to push. (Pass the JS literal as a string,
                            e.g. "`<td>NEW</td>`" or '"<td>NEW</td>"'.)

    Returns:
        str: The modified HTML.
    """
    # This pattern finds:
    #  1. The function signature and opening brace.
    #  2. All content (non-greedily) until we hit a newline containing a return statement.
    #  3. Captures the newline and leading whitespace (indent) of the return statement.
    #  4. Captures the rest of the return line.
    pattern = (
        r"("
        + re.escape(js_function_name)
        + r"\s*\([^)]*\)\s*\{)"  # group1: function header
        r"([\s\S]*?)"  # group2: code before return
        r"(\n(\s*)return\s+[^;]+;)"  # group3: newline+return line, group4: indent
    )

    def replacer(match):
        header = match.group(1)
        before_return = match.group(2)
        return_line = match.group(3)
        indent = match.group(4)
        # Create the new push statement. We add a newline plus the same indent.
        # The resulting line will be, e.g.,
        # "        rowComponents.push(`<td>NEW</td>`);"
        injection = f"\n{indent}{js_array_name}.push({js_new_value});"
        return header + before_return + injection + return_line

    # Use re.sub to replace only the first occurrence of the function
    new_html, count = re.subn(
        pattern, replacer, html_str, count=1, flags=re.MULTILINE | re.DOTALL
    )
    return new_html


def update_my_app_name_test_create(ctx: AnyContext):
    snake_module_name = to_snake_case(ctx.input.module)
    snake_entity_name = to_snake_case(ctx.input.entity)
    test_file_path = os.path.join(
        APP_DIR,
        "test",
        snake_module_name,
        snake_entity_name,
        f"test_create_{snake_entity_name}.py",
    )
    existing_code = read_file(test_file_path)
    new_code = existing_code
    # TODO: update test
    write_file(test_file_path, new_code)


def update_my_app_name_test_read(ctx: AnyContext):
    snake_module_name = to_snake_case(ctx.input.module)
    snake_entity_name = to_snake_case(ctx.input.entity)
    test_file_path = os.path.join(
        APP_DIR,
        "test",
        snake_module_name,
        snake_entity_name,
        f"test_read_{snake_entity_name}.py",
    )
    existing_code = read_file(test_file_path)
    new_code = existing_code
    # TODO: update test
    write_file(test_file_path, new_code)


def update_my_app_name_test_update(ctx: AnyContext):
    snake_module_name = to_snake_case(ctx.input.module)
    snake_entity_name = to_snake_case(ctx.input.entity)
    test_file_path = os.path.join(
        APP_DIR,
        "test",
        snake_module_name,
        snake_entity_name,
        f"test_update_{snake_entity_name}.py",
    )
    existing_code = read_file(test_file_path)
    new_code = existing_code
    # TODO: update test
    write_file(test_file_path, new_code)


def update_my_app_name_test_delete(ctx: AnyContext):
    snake_module_name = to_snake_case(ctx.input.module)
    snake_entity_name = to_snake_case(ctx.input.entity)
    test_file_path = os.path.join(
        APP_DIR,
        "test",
        snake_module_name,
        snake_entity_name,
        f"test_delete_{snake_entity_name}.py",
    )
    existing_code = read_file(test_file_path)
    new_code = existing_code
    # TODO: update test
    write_file(test_file_path, new_code)



import libcst as cst

from zrb.util.codemod.modification_mode import APPEND, PREPEND, REPLACE


def replace_class_code(original_code: str, class_name: str, new_code: str) -> str:
    return _modify_code(original_code, class_name, new_code, REPLACE)


def prepend_code_to_class(original_code: str, class_name: str, new_code: str) -> str:
    return _modify_code(original_code, class_name, new_code, PREPEND)


def append_code_to_class(original_code: str, class_name: str, new_code: str) -> str:
    return _modify_code(original_code, class_name, new_code, APPEND)


def _modify_code(original_code: str, class_name: str, new_code: str, mode: int) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize transformer with the class name and method code
    transformer = _ClassCodeModifier(class_name, new_code, mode)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Check if the class was found
    if not transformer.class_found:
        raise ValueError(f"Class {class_name} not found in the provided code.")
    # Return the modified code
    return modified_module.code


class _ClassCodeModifier(cst.CSTTransformer):
    def __init__(self, class_name: str, new_code: str, mode: int):
        self.class_name = class_name
        self.new_code = cst.parse_module(new_code).body
        self.class_found = False
        self.mode = mode

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        # Check if this is the target class
        if original_node.name.value == self.class_name:
            self.class_found = True
            if self.mode == REPLACE:
                new_body = updated_node.body.with_changes(body=tuple(self.new_code))
                return updated_node.with_changes(body=new_body)
            if self.mode == PREPEND:
                new_body = updated_node.body.with_changes(
                    body=tuple(self.new_code) + updated_node.body.body
                )
                return updated_node.with_changes(body=new_body)
            if self.mode == APPEND:
                new_body = updated_node.body.with_changes(
                    body=updated_node.body.body + tuple(self.new_code)
                )
                return updated_node.with_changes(body=new_body)
        return updated_node



import os
import sys

from zrb.config import INIT_MODULES, INIT_SCRIPTS
from zrb.runner.cli import cli
from zrb.util.cli.style import stylize_error, stylize_warning
from zrb.util.group import NodeNotFoundError
from zrb.util.load import load_file, load_module


def serve_cli():
    try:
        # load init modules
        for init_module in INIT_MODULES:
            load_module(init_module)
        zrb_init_path_list = _get_zrb_init_path_list()
        # load init scripts
        for init_script in INIT_SCRIPTS:
            abs_init_script = os.path.abspath(init_script)
            if abs_init_script not in zrb_init_path_list:
                load_file(abs_init_script, -1)
        # load zrb init
        for zrb_init_path in zrb_init_path_list:
            load_file(zrb_init_path)
        # run the CLI
        cli.run(sys.argv[1:])
    except KeyboardInterrupt:
        print(stylize_warning("\nStopped"), file=sys.stderr)
        sys.exit(1)
    except RuntimeError as e:
        if f"{e}".lower() != "event loop is closed":
            raise e
        sys.exit(1)
    except NodeNotFoundError as e:
        print(stylize_error(f"{e}"), file=sys.stderr)
        sys.exit(1)


def _get_zrb_init_path_list() -> list[str]:
    current_path = os.path.abspath(os.getcwd())
    dir_path_list = [current_path]
    while current_path != os.path.dirname(current_path):  # Stop at root
        current_path = os.path.dirname(current_path)
        dir_path_list.append(current_path)
    zrb_init_path_list = []
    for current_path in dir_path_list[::-1]:
        zrb_init_path = os.path.join(current_path, "zrb_init.py")
        if os.path.isfile(zrb_init_path):
            zrb_init_path_list.append(zrb_init_path)
    return zrb_init_path_list



import json
from typing import Annotated, Literal


def get_current_location() -> (
    Annotated[str, "JSON string representing latitude and longitude"]
):  # noqa
    """Get the user's current location. This function take no argument."""
    import requests

    return json.dumps(requests.get("http://ip-api.com/json?fields=lat,lon").json())


def get_current_weather(
    latitude: float,
    longitude: float,
    temperature_unit: Literal["celsius", "fahrenheit"],
) -> str:
    """Get the current weather in a given location."""
    import requests

    resp = requests.get(
        "https://api.open-meteo.com/v1/forecast",
        params={
            "latitude": latitude,
            "longitude": longitude,
            "temperature_unit": temperature_unit,
            "current_weather": True,
        },
    )
    return json.dumps(resp.json())



from abc import ABC, abstractmethod
from typing import Any

from zrb.context.any_shared_context import AnySharedContext


class AnyCallback(ABC):
    @abstractmethod
    async def async_run(self, parent_ctx: AnySharedContext) -> Any:
        pass



# MyEntity routes


@app.get("/my-module/my-entities", include_in_schema=False)
def my_entities_crud_ui(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    page: int = 1,
    page_size: int = 10,
    sort: str | None = None,
    filter: str | None = None,
):
    if not current_user.has_permission("my-entity:read"):
        return render_error(error_message="Access denied", status_code=403)
    return render_content(
        view_path=os.path.join("my-module", "my-entity.html"),
        current_user=current_user,
        page_name="my-module.my-entity",
        page=page,
        page_size=page_size,
        sort=sort,
        filter=filter,
        allow_create=current_user.has_permission("my-entity:create"),
        allow_update=current_user.has_permission("my-entity:update"),
        allow_delete=current_user.has_permission("my-entity:delete"),
    )


@app.get("/api/v1/my-entities", response_model=MultipleMyEntityResponse)
async def get_my_entities(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    page: int = 1,
    page_size: int = 10,
    sort: str | None = None,
    filter: str | None = None,
) -> MultipleMyEntityResponse:
    if not current_user.has_permission("my-entity:read"):
        raise ForbiddenError("Access denied")
    return await my_module_client.get_my_entities(
        page=page, page_size=page_size, sort=sort, filter=filter
    )


@app.get("/api/v1/my-entities/{my_entity_id}", response_model=MyEntityResponse)
async def get_my_entity_by_id(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    my_entity_id: str,
) -> MyEntityResponse:
    if not current_user.has_permission("my-entity:read"):
        raise ForbiddenError("Access denied")
    return await my_module_client.get_my_entity_by_id(my_entity_id)


@app.post(
    "/api/v1/my-entities/bulk",
    response_model=list[MyEntityResponse],
)
async def create_my_entity_bulk(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    data: list[MyEntityCreate],
) -> list[MyEntityResponse]:
    if not current_user.has_permission("my-entity:create"):
        raise ForbiddenError("Access denied")
    return await my_module_client.create_my_entity_bulk(
        [row.with_audit(created_by=current_user.id) for row in data]
    )


@app.post(
    "/api/v1/my-entities",
    response_model=MyEntityResponse,
)
async def create_my_entity(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    data: MyEntityCreate,
) -> MyEntityResponse:
    if not current_user.has_permission("my-entity:create"):
        raise ForbiddenError("Access denied")
    return await my_module_client.create_my_entity(
        data.with_audit(created_by=current_user.id)
    )


@app.put(
    "/api/v1/my-entities/bulk",
    response_model=list[MyEntityResponse],
)
async def update_my_entity_bulk(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    my_entity_ids: list[str],
    data: MyEntityUpdate,
) -> list[MyEntityResponse]:
    if not current_user.has_permission("my-entity:update"):
        raise ForbiddenError("Access denied")
    return await my_module_client.update_my_entity_bulk(
        my_entity_ids, data.with_audit(updated_by=current_user.id)
    )


@app.put(
    "/api/v1/my-entities/{my_entity_id}",
    response_model=MyEntityResponse,
)
async def update_my_entity(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    my_entity_id: str,
    data: MyEntityUpdate,
) -> MyEntityResponse:
    if not current_user.has_permission("my-entity:update"):
        raise ForbiddenError("Access denied")
    return await my_module_client.update_my_entity(
        my_entity_id, data.with_audit(updated_by=current_user.id)
    )


@app.delete(
    "/api/v1/my-entities/bulk",
    response_model=list[MyEntityResponse],
)
async def delete_my_entity_bulk(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    my_entity_ids: list[str],
) -> list[MyEntityResponse]:
    if not current_user.has_permission("my-entity:delete"):
        raise ForbiddenError("Access denied")
    return await my_module_client.delete_my_entity_bulk(
        my_entity_ids, deleted_by=current_user.id
    )


@app.delete(
    "/api/v1/my-entities/{my_entity_id}",
    response_model=MyEntityResponse,
)
async def delete_my_entity(
    current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
    my_entity_id: str,
) -> MyEntityResponse:
    if not current_user.has_permission("my-entity:delete"):
        raise ForbiddenError("Access denied")
    return await my_module_client.delete_my_entity(
        my_entity_id, deleted_by=current_user.id
    )



import libcst as cst

from zrb.util.codemod.modification_mode import APPEND, PREPEND, REPLACE


def replace_parent_class(
    original_code: str, class_name: str, parent_class_name: str
) -> str:
    return _modify_parent_class(original_code, class_name, parent_class_name, REPLACE)


def append_parent_class(
    original_code: str, class_name: str, parent_class_name: str
) -> str:
    return _modify_parent_class(original_code, class_name, parent_class_name, APPEND)


def prepend_parent_class(
    original_code: str, class_name: str, parent_class_name: str
) -> str:
    return _modify_parent_class(original_code, class_name, parent_class_name, PREPEND)


def _modify_parent_class(
    original_code: str, class_name: str, parent_class_name: str, mode: int
) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize transformer with the class name and parent class name
    transformer = _ParentClassAdder(class_name, parent_class_name, mode)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Check if the class was found
    if not transformer.class_found:
        raise ValueError(f"Class {class_name} not found in the provided code.")
    # Return the modified code
    return modified_module.code


class _ParentClassAdder(cst.CSTTransformer):
    def __init__(self, class_name: str, parent_class_name: str, mode: int):
        self.class_name = class_name
        self.parent_class_name = parent_class_name
        self.class_found = False
        self.mode = mode

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        # Check if this is the target class
        if original_node.name.value == self.class_name:
            self.class_found = True
            if self.mode == REPLACE:
                new_bases = (cst.Arg(value=cst.Name(self.parent_class_name)),)
                return updated_node.with_changes(bases=new_bases)
            if self.mode == PREPEND:
                new_bases = (
                    cst.Arg(value=cst.Name(self.parent_class_name)),
                    *updated_node.bases,
                )
                return updated_node.with_changes(bases=new_bases)
            if self.mode == APPEND:
                new_bases = (
                    *updated_node.bases,
                    cst.Arg(value=cst.Name(self.parent_class_name)),
                )
                return updated_node.with_changes(bases=new_bases)
        return updated_node



from typing import Dict

from fastapi import HTTPException


class NotFoundError(HTTPException):
    def __init__(self, message: str, headers: Dict[str, str] | None = None) -> None:
        super().__init__(404, {"message": message}, headers)


class ForbiddenError(HTTPException):
    def __init__(self, message: str, headers: Dict[str, str] | None = None) -> None:
        super().__init__(403, {"message": message}, headers)


class UnauthorizedError(HTTPException):
    def __init__(self, message: str, headers: Dict[str, str] | None = None) -> None:
        super().__init__(401, {"message": message}, headers)


class InvalidValueError(HTTPException):
    def __init__(self, message: str, headers: Dict[str, str] | None = None) -> None:
        super().__init__(422, {"message": message}, headers)


class InternalServerError(HTTPException):
    def __init__(self, message: str, headers: Dict[str, str] | None = None) -> None:
        super().__init__(500, {"message": message}, headers)


class ClientAPIError(HTTPException):
    def __init__(
        self, status_code: int, message: str, headers: Dict[str, str] | None = None
    ) -> None:
        super().__init__(status_code, {"message": message}, headers)



import os

from pydantic_ai.models import Model
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers import Provider
from pydantic_ai.providers.openai import OpenAIProvider

DEFAULT_SYSTEM_PROMPT = """
You have access to tools.
Your goal is to answer user queries accurately.
Follow these instructions precisely:
1. ALWAYS use available tools to gather information BEFORE asking the user questions
2. For tools that require arguments: provide arguments in valid JSON format
3. For tools with no args: call the tool without args. Do NOT pass "" or {}.
4. NEVER pass arguments to tools that don't accept parameters
5. NEVER ask users for information obtainable through tools
6. Use tools in a logical sequence until you have sufficient information
7. If a tool call fails, check if you're passing arguments in the correct format
8. Only after exhausting relevant tools should you request clarification
""".strip()


class LLMConfig:

    def __init__(
        self,
        default_model_name: str | None = None,
        default_base_url: str | None = None,
        default_api_key: str | None = None,
        default_system_prompt: str | None = None,
    ):
        self._model_name = (
            default_model_name
            if default_model_name is not None
            else os.getenv("ZRB_LLM_MODEL", None)
        )
        self._model_base_url = (
            default_base_url
            if default_base_url is not None
            else os.getenv("ZRB_LLM_BASE_URL", None)
        )
        self._model_api_key = (
            default_api_key
            if default_api_key is not None
            else os.getenv("ZRB_LLM_API_KEY", None)
        )
        self._system_prompt = (
            default_system_prompt
            if default_system_prompt is not None
            else os.getenv("ZRB_LLM_SYSTEM_PROMPT", None)
        )
        self._default_provider = None
        self._default_model = None

    def _get_model_name(self) -> str | None:
        return self._model_name if self._model_name is not None else None

    def get_default_model_provider(self) -> Provider | str:
        if self._default_provider is not None:
            return self._default_provider
        if self._model_base_url is None and self._model_api_key is None:
            return "openai"
        return OpenAIProvider(
            base_url=self._model_base_url, api_key=self._model_api_key
        )

    def get_default_system_prompt(self) -> str:
        if self._system_prompt is not None:
            return self._system_prompt
        return DEFAULT_SYSTEM_PROMPT

    def get_default_model(self) -> Model | str | None:
        if self._default_model is not None:
            return self._default_model
        model_name = self._get_model_name()
        if model_name is None:
            return None
        return OpenAIModel(
            model_name=model_name,
            provider=self.get_default_model_provider(),
        )

    def set_default_system_prompt(self, system_prompt: str):
        self._system_prompt = system_prompt

    def set_default_model_name(self, model_name: str):
        self._model_name = model_name

    def set_default_model_api_key(self, model_api_key: str):
        self._model_api_key = model_api_key

    def set_default_model_base_url(self, model_base_url: str):
        self._model_base_url = model_base_url

    def set_default_provider(self, provider: Provider | str):
        self._default_provider = provider

    def set_default_model(self, model: Model | str | None):
        self._default_model = model


llm_config = LLMConfig()



import os

from fastapi.testclient import TestClient
from my_app_name.main import app
from my_app_name.module.gateway.util.view import render_content


def test_homepage():
    client = TestClient(app, base_url="http://localhost")
    response = client.get("/")
    assert response.status_code == 200
    view_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "module", "gateway", "view"
    )
    assert response.text == render_content("homepage.html").body.decode("utf-8")



import os
from typing import Annotated

from fastapi import Depends, FastAPI, Request, Response
from fastapi.security import OAuth2PasswordRequestForm
from my_app_name.common.error import ForbiddenError, NotFoundError
from my_app_name.module.auth.client.auth_client_factory import auth_client
from my_app_name.module.gateway.util.auth import (
    get_current_user,
    get_refresh_token,
    set_user_session_cookie,
    unset_user_session_cookie,
)
from my_app_name.module.gateway.util.view import render_content, render_error
from my_app_name.schema.permission import (
    MultiplePermissionResponse,
    PermissionCreate,
    PermissionResponse,
    PermissionUpdate,
)
from my_app_name.schema.role import (
    MultipleRoleResponse,
    RoleCreateWithPermissions,
    RoleResponse,
    RoleUpdateWithPermissions,
)
from my_app_name.schema.user import (
    AuthUserResponse,
    MultipleUserResponse,
    UserCreateWithRoles,
    UserCredentials,
    UserResponse,
    UserSessionResponse,
    UserUpdateWithRoles,
)


def serve_auth_route(app: FastAPI):

    @app.post("/api/v1/user-sessions", response_model=UserSessionResponse)
    async def create_user_session(
        response: Response, form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
    ) -> UserSessionResponse:
        user_session = await auth_client.create_user_session(
            UserCredentials(
                username=form_data.username,
                password=form_data.password,
            )
        )
        set_user_session_cookie(response, user_session)
        return user_session

    @app.put("/api/v1/user-sessions", response_model=UserSessionResponse)
    async def update_user_session(
        request: Request, response: Response, refresh_token: str | None = None
    ) -> UserSessionResponse:
        actual_refresh_token = get_refresh_token(request, refresh_token)
        if actual_refresh_token is None:
            raise ForbiddenError("Refresh token needed")
        try:
            user_session = await auth_client.update_user_session(actual_refresh_token)
        except NotFoundError:
            raise ForbiddenError("Session not found")
        set_user_session_cookie(response, user_session)
        return user_session

    @app.delete("/api/v1/user-sessions", response_model=UserSessionResponse)
    async def delete_user_session(
        request: Request, response: Response, refresh_token: str | None = None
    ) -> UserSessionResponse:
        try:
            actual_refresh_token = get_refresh_token(request, refresh_token)
            if actual_refresh_token is None:
                raise ForbiddenError("Refresh token needed")
            user_session = await auth_client.delete_user_session(actual_refresh_token)
            return user_session
        finally:
            unset_user_session_cookie(response)

    # Permission routes

    @app.get("/auth/permissions", include_in_schema=False)
    def permissions_crud_ui(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ):
        if not current_user.has_permission("permission:read"):
            return render_error(error_message="Access denied", status_code=403)
        return render_content(
            view_path=os.path.join("auth", "permission.html"),
            current_user=current_user,
            page_name="auth.permission",
            page=page,
            page_size=page_size,
            sort=sort,
            filter=filter,
            allow_create=current_user.has_permission("permission:create"),
            allow_update=current_user.has_permission("permission:update"),
            allow_delete=current_user.has_permission("permission:delete"),
        )

    @app.get("/api/v1/permissions", response_model=MultiplePermissionResponse)
    async def get_permissions(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ) -> MultiplePermissionResponse:
        if not current_user.has_permission("permission:read"):
            raise ForbiddenError("Access denied")
        return await auth_client.get_permissions(
            page=page, page_size=page_size, sort=sort, filter=filter
        )

    @app.get("/api/v1/permissions/{permission_id}", response_model=PermissionResponse)
    async def get_permission_by_id(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        permission_id: str,
    ) -> PermissionResponse:
        if not current_user.has_permission("permission:read"):
            raise ForbiddenError("Access denied")
        return await auth_client.get_permission_by_id(permission_id)

    @app.post(
        "/api/v1/permissions/bulk",
        response_model=list[PermissionResponse],
    )
    async def create_permission_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        data: list[PermissionCreate],
    ) -> list[PermissionResponse]:
        if not current_user.has_permission("permission:create"):
            raise ForbiddenError("Access denied")
        return await auth_client.create_permission_bulk(
            [row.with_audit(created_by=current_user.id) for row in data]
        )

    @app.post(
        "/api/v1/permissions",
        response_model=PermissionResponse,
    )
    async def create_permission(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        data: PermissionCreate,
    ) -> PermissionResponse:
        if not current_user.has_permission("permission:create"):
            raise ForbiddenError("Access denied")
        return await auth_client.create_permission(
            data.with_audit(created_by=current_user.id)
        )

    @app.put(
        "/api/v1/permissions/bulk",
        response_model=list[PermissionResponse],
    )
    async def update_permission_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        permission_ids: list[str],
        data: PermissionUpdate,
    ) -> list[PermissionResponse]:
        if not current_user.has_permission("permission:update"):
            raise ForbiddenError("Access denied")
        return await auth_client.update_permission_bulk(
            permission_ids, data.with_audit(updated_by=current_user.id)
        )

    @app.put(
        "/api/v1/permissions/{permission_id}",
        response_model=PermissionResponse,
    )
    async def update_permission(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        permission_id: str,
        data: PermissionUpdate,
    ) -> PermissionResponse:
        if not current_user.has_permission("permission:update"):
            raise ForbiddenError("Access denied")
        return await auth_client.update_permission(
            permission_id, data.with_audit(updated_by=current_user.id)
        )

    @app.delete(
        "/api/v1/permissions/bulk",
        response_model=list[PermissionResponse],
    )
    async def delete_permission_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        permission_ids: list[str],
    ) -> list[PermissionResponse]:
        if not current_user.has_permission("permission:delete"):
            raise ForbiddenError("Access denied")
        return await auth_client.delete_permission_bulk(
            permission_ids, deleted_by=current_user.id
        )

    @app.delete(
        "/api/v1/permissions/{permission_id}",
        response_model=PermissionResponse,
    )
    async def delete_permission(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        permission_id: str,
    ) -> PermissionResponse:
        if not current_user.has_permission("permission:delete"):
            raise ForbiddenError("Access denied")
        return await auth_client.delete_permission(
            permission_id, deleted_by=current_user.id
        )

    # Role routes

    @app.get("/auth/roles", include_in_schema=False)
    def roles_crud_ui(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ):
        if not current_user.has_permission("role:read"):
            return render_error(error_message="Access denied", status_code=403)
        return render_content(
            view_path=os.path.join("auth", "role.html"),
            current_user=current_user,
            page_name="auth.role",
            page=page,
            page_size=page_size,
            sort=sort,
            filter=filter,
            allow_create=current_user.has_permission("role:create"),
            allow_update=current_user.has_permission("role:update"),
            allow_delete=current_user.has_permission("role:delete"),
        )

    @app.get("/api/v1/roles", response_model=MultipleRoleResponse)
    async def get_roles(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ) -> MultipleRoleResponse:
        if not current_user.has_permission("role:read"):
            raise ForbiddenError("Access denied")
        return await auth_client.get_roles(
            page=page, page_size=page_size, sort=sort, filter=filter
        )

    @app.get("/api/v1/roles/{role_id}", response_model=RoleResponse)
    async def get_role_by_id(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        role_id: str,
    ) -> RoleResponse:
        if not current_user.has_permission("role:read"):
            raise ForbiddenError("Access denied")
        return await auth_client.get_role_by_id(role_id)

    @app.post(
        "/api/v1/roles/bulk",
        response_model=list[RoleResponse],
    )
    async def create_role_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        data: list[RoleCreateWithPermissions],
    ) -> list[RoleResponse]:
        if not current_user.has_permission("role:create"):
            raise ForbiddenError("Access denied")
        return await auth_client.create_role_bulk(
            [row.with_audit(created_by=current_user.id) for row in data]
        )

    @app.post(
        "/api/v1/roles",
        response_model=RoleResponse,
    )
    async def create_role(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        data: RoleCreateWithPermissions,
    ) -> RoleResponse:
        if not current_user.has_permission("role:create"):
            raise ForbiddenError("Access denied")
        return await auth_client.create_role(
            data.with_audit(created_by=current_user.id)
        )

    @app.put(
        "/api/v1/roles/bulk",
        response_model=list[RoleResponse],
    )
    async def update_role_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        role_ids: list[str],
        data: RoleUpdateWithPermissions,
    ) -> list[RoleResponse]:
        if not current_user.has_permission("role:update"):
            raise ForbiddenError("Access denied")
        return await auth_client.update_role_bulk(
            role_ids, data.with_audit(updated_by=current_user.id)
        )

    @app.put(
        "/api/v1/roles/{role_id}",
        response_model=RoleResponse,
    )
    async def update_role(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        role_id: str,
        data: RoleUpdateWithPermissions,
    ) -> RoleResponse:
        if not current_user.has_permission("role:update"):
            raise ForbiddenError("Access denied")
        return await auth_client.update_role(
            role_id, data.with_audit(updated_by=current_user.id)
        )

    @app.delete(
        "/api/v1/roles/bulk",
        response_model=list[RoleResponse],
    )
    async def delete_role_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        role_ids: list[str],
    ) -> list[RoleResponse]:
        if not current_user.has_permission("role:delete"):
            raise ForbiddenError("Access denied")
        return await auth_client.delete_role_bulk(role_ids, deleted_by=current_user.id)

    @app.delete(
        "/api/v1/roles/{role_id}",
        response_model=RoleResponse,
    )
    async def delete_role(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        role_id: str,
    ) -> RoleResponse:
        if not current_user.has_permission("role:delete"):
            raise ForbiddenError("Access denied")
        return await auth_client.delete_role(role_id, deleted_by=current_user.id)

    # User routes

    @app.get("/auth/users", include_in_schema=False)
    def users_crud_ui(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ):
        if not current_user.has_permission("user:read"):
            return render_error(error_message="Access denied", status_code=403)
        return render_content(
            view_path=os.path.join("auth", "user.html"),
            current_user=current_user,
            page_name="auth.user",
            page=page,
            page_size=page_size,
            sort=sort,
            filter=filter,
            allow_create=current_user.has_permission("user:create"),
            allow_update=current_user.has_permission("user:update"),
            allow_delete=current_user.has_permission("user:delete"),
        )

    @app.get("/api/v1/users", response_model=MultipleUserResponse)
    async def get_users(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        page: int = 1,
        page_size: int = 10,
        sort: str | None = None,
        filter: str | None = None,
    ) -> MultipleUserResponse:
        if not current_user.has_permission("user:read"):
            raise ForbiddenError("Access denied")
        return await auth_client.get_users(
            page=page, page_size=page_size, sort=sort, filter=filter
        )

    @app.get("/api/v1/users/{user_id}", response_model=UserResponse)
    async def get_user_by_id(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        user_id: str,
    ) -> UserResponse:
        if not current_user.has_permission("user:read"):
            raise ForbiddenError("Access denied")
        return await auth_client.get_user_by_id(user_id)

    @app.post(
        "/api/v1/users/bulk",
        response_model=list[UserResponse],
    )
    async def create_user_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        data: list[UserCreateWithRoles],
    ) -> list[UserResponse]:
        if not current_user.has_permission("user:create"):
            raise ForbiddenError("Access denied")
        return await auth_client.create_user_bulk(
            [row.with_audit(created_by=current_user.id) for row in data]
        )

    @app.post(
        "/api/v1/users",
        response_model=UserResponse,
    )
    async def create_user(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        data: UserCreateWithRoles,
    ) -> UserResponse:
        if not current_user.has_permission("user:create"):
            raise ForbiddenError("Access denied")
        return await auth_client.create_user(
            data.with_audit(created_by=current_user.id)
        )

    @app.put(
        "/api/v1/users/bulk",
        response_model=list[UserResponse],
    )
    async def update_user_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        user_ids: list[str],
        data: UserUpdateWithRoles,
    ) -> list[UserResponse]:
        if not current_user.has_permission("user:update"):
            raise ForbiddenError("Access denied")
        return await auth_client.update_user_bulk(
            user_ids, data.with_audit(updated_by=current_user.id)
        )

    @app.put(
        "/api/v1/users/{user_id}",
        response_model=UserResponse,
    )
    async def update_user(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        user_id: str,
        data: UserUpdateWithRoles,
    ) -> UserResponse:
        if not current_user.has_permission("user:update"):
            raise ForbiddenError("Access denied")
        return await auth_client.update_user(
            user_id, data.with_audit(updated_by=current_user.id)
        )

    @app.delete(
        "/api/v1/users/bulk",
        response_model=list[UserResponse],
    )
    async def delete_user_bulk(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        user_ids: list[str],
    ) -> list[UserResponse]:
        if not current_user.has_permission("user:delete"):
            raise ForbiddenError("Access denied")
        return await auth_client.delete_user_bulk(user_ids, deleted_by=current_user.id)

    @app.delete(
        "/api/v1/users/{user_id}",
        response_model=UserResponse,
    )
    async def delete_user(
        current_user: Annotated[AuthUserResponse, Depends(get_current_user)],
        user_id: str,
    ) -> UserResponse:
        if not current_user.has_permission("user:delete"):
            raise ForbiddenError("Access denied")
        return await auth_client.delete_user(user_id, deleted_by=current_user.id)



from dotenv import dotenv_values

from zrb.attr.type import StrAttr
from zrb.context.shared_context import SharedContext
from zrb.env.env_map import EnvMap
from zrb.util.attr import get_str_attr


class EnvFile(EnvMap):
    def __init__(
        self,
        path: StrAttr,
        auto_render: bool = True,
        link_to_os: bool = True,
        os_prefix: str | None = None,
    ):
        super().__init__(
            vars={}, auto_render=auto_render, link_to_os=link_to_os, os_prefix=os_prefix
        )
        self._file_path = path

    def _get_env_map(self, shared_ctx: SharedContext) -> dict[str, str]:
        file_path = get_str_attr(shared_ctx, self._file_path, ".env", self._auto_render)
        return dotenv_values(file_path)



from zrb.config import (
    WEB_ACCESS_TOKEN_COOKIE_NAME,
    WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES,
    WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES,
    WEB_ENABLE_AUTH,
    WEB_GUEST_USERNAME,
    WEB_HTTP_PORT,
    WEB_REFRESH_TOKEN_COOKIE_NAME,
    WEB_SECRET_KEY,
    WEB_SUPER_ADMIN_PASSWORD,
    WEB_SUPER_ADMIN_USERNAME,
)
from zrb.runner.web_config.config import WebConfig

web_config = WebConfig(
    port=WEB_HTTP_PORT,
    secret_key=WEB_SECRET_KEY,
    access_token_expire_minutes=WEB_AUTH_ACCESS_TOKEN_EXPIRE_MINUTES,
    refresh_token_expire_minutes=WEB_AUTH_REFRESH_TOKEN_EXPIRE_MINUTES,
    access_token_cookie_name=WEB_ACCESS_TOKEN_COOKIE_NAME,
    refresh_token_cookie_name=WEB_REFRESH_TOKEN_COOKIE_NAME,
    enable_auth=WEB_ENABLE_AUTH,
    super_admin_username=WEB_SUPER_ADMIN_USERNAME,
    super_admin_password=WEB_SUPER_ADMIN_PASSWORD,
    guest_username=WEB_GUEST_USERNAME,
)



import random
import string

PREFIXES = [
    "bold",
    "calm",
    "dark",
    "deep",
    "fast",
    "firm",
    "glad",
    "grey",
    "hard",
    "high",
    "kind",
    "late",
    "lean",
    "long",
    "loud",
    "mild",
    "neat",
    "pure",
    "rare",
    "rich",
    "safe",
    "slow",
    "soft",
    "tall",
    "thin",
    "trim",
    "vast",
    "warm",
    "weak",
    "wild",
]
SUFFIXES = [
    "arch",
    "area",
    "atom",
    "base",
    "beam",
    "bell",
    "bolt",
    "bone",
    "bulk",
    "bush",
    "cell",
    "chip",
    "clay",
    "coal",
    "coil",
    "cone",
    "cube",
    "disk",
    "dust",
    "face",
    "film",
    "foam",
    "frog",
    "fuel",
    "gate",
    "gear",
    "hall",
    "hand",
    "horn",
    "leaf",
]


def get_random_name(
    separator: str = "-", add_random_digit: bool = True, digit_count: int = 4
) -> str:
    prefix = random.choice(PREFIXES)
    suffix = random.choice(SUFFIXES)
    parts = [prefix, suffix]
    if add_random_digit:
        random_digit = "".join(random.choices(string.digits, k=digit_count))
        parts.append(random_digit)
    return separator.join(parts)



from fastapi import FastAPI
from my_app_name.common.app_factory import app
from my_app_name.common.schema import BasicResponse
from my_app_name.config import APP_MAIN_MODULE, APP_MODE, APP_MODULES
from my_app_name.module.auth.service.permission.permission_service_factory import (
    permission_service,
)
from my_app_name.module.auth.service.role.role_service_factory import role_service
from my_app_name.module.auth.service.user.user_service_factory import user_service


def serve_route(app: FastAPI):
    if APP_MODE != "microservices" or "auth" not in APP_MODULES:
        return
    if APP_MAIN_MODULE == "auth":
        _serve_health_check(app)
        _serve_readiness_check(app)

    permission_service.serve_route(app)
    role_service.serve_route(app)
    user_service.serve_route(app)


def _serve_health_check(app: FastAPI):
    @app.api_route("/health", methods=["GET", "HEAD"], response_model=BasicResponse)
    async def health():
        """
        Microservice's health check
        """
        return BasicResponse(message="ok")


def _serve_readiness_check(app: FastAPI):
    @app.api_route("/readiness", methods=["GET", "HEAD"], response_model=BasicResponse)
    async def readiness():
        """
        Microservice's readiness check
        """
        return BasicResponse(message="ok")


serve_route(app)



import libcst as cst

from zrb.util.codemod.modification_mode import APPEND, PREPEND, REPLACE


def replace_function_call_param(
    original_code: str, function_name: str, new_param: str
) -> str:
    return _modify_function_call(original_code, function_name, new_param, REPLACE)


def prepend_param_to_function_call(
    original_code: str, function_name: str, new_param: str
) -> str:
    return _modify_function_call(original_code, function_name, new_param, PREPEND)


def append_param_to_function_call(
    original_code: str, function_name: str, new_param: str
) -> str:
    return _modify_function_call(original_code, function_name, new_param, APPEND)


def _modify_function_call(
    original_code: str, function_name: str, new_param: str, mode: int
) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize the transformer with the necessary information
    transformer = _FunctionCallParamModifier(function_name, new_param, mode)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Error handling: raise an error if the function call is not found
    if not transformer.param_added:
        raise ValueError(
            f"Function call to {function_name} not found in the provided code."
        )
    # Return the modified code
    return modified_module.code


class _FunctionCallParamModifier(cst.CSTTransformer):
    def __init__(self, func_name: str, new_param: str, mode: int):
        self.func_name = func_name
        # Parse the new parameter to ensure it’s a valid CST node
        self.new_param = cst.parse_expression(new_param)
        self.param_added = False
        self.mode = mode

    def leave_Call(self, original_node: cst.Call, updated_node: cst.Call) -> cst.Call:
        # Check if the function call name matches the target function
        if (
            isinstance(original_node.func, cst.Name)
            and original_node.func.value == self.func_name
        ):
            if self.mode == REPLACE:
                new_args = (cst.Arg(value=self.new_param),)
                self.param_added = True
                return updated_node.with_changes(args=new_args)
            if self.mode == PREPEND:
                new_args = (cst.Arg(value=self.new_param),) + updated_node.args
                self.param_added = True
                return updated_node.with_changes(args=new_args)
            if self.mode == APPEND:
                new_args = updated_node.args + (cst.Arg(value=self.new_param),)
                self.param_added = True
                return updated_node.with_changes(args=new_args)
        return updated_node



import fnmatch
import os

from zrb.util.file import read_file, write_file

_INCLUDED_PATTERNS: list[str] = [
    "*.py",  # Python
    "*.go",  # Go
    "*.rs",  # Rust
    "*.js",  # JavaScript
    "*.ts",  # TypeScript
    "*.java",  # Java
    "*.c",  # C
    "*.cpp",  # C++
    "*.cc",  # Alternative C++ extension
    "*.cxx",  # Alternative C++ extension
    "*.rb",  # Ruby
    "*.swift",  # Swift
    "*.kt",  # Kotlin
    "*.php",  # PHP
    "*.pl",  # Perl / Prolog
    "*.pm",  # Perl module
    "*.sh",  # Shell
    "*.bat",  # Batch
    "*.ps1",  # PowerShell
    "*.R",  # R (capital)
    "*.r",  # R (lowercase)
    "*.scala",  # Scala
    "*.hs",  # Haskell
    "*.cs",  # C#
    "*.fs",  # F#
    "*.ex",  # Elixir
    "*.exs",  # Elixir script
    "*.erl",  # Erlang
    "*.hrl",  # Erlang header
    "*.dart",  # Dart
    "*.m",  # Objective-C / Matlab (note: conflicts may arise)
    "*.mm",  # Objective-C++
    "*.lua",  # Lua
    "*.jl",  # Julia
    "*.groovy",  # Groovy
    "*.clj",  # Clojure
    "*.cljs",  # ClojureScript
    "*.cljc",  # Clojure common
    "*.vb",  # Visual Basic
    "*.f90",  # Fortran
    "*.f95",  # Fortran
    "*.adb",  # Ada
    "*.ads",  # Ada specification
    "*.pas",  # Pascal
    "*.pp",  # Pascal
    "*.ml",  # OCaml
    "*.mli",  # OCaml interface
    "*.nim",  # Nim
    "*.rkt",  # Racket
    "*.d",  # D
    "*.lisp",  # Common Lisp
    "*.lsp",  # Lisp variant
    "*.cl",  # Common Lisp
    "*.scm",  # Scheme
    "*.st",  # Smalltalk
    "*.vhd",  # VHDL
    "*.vhdl",  # VHDL
    "*.v",  # Verilog
    "*.asm",  # Assembly
    "*.s",  # Assembly (alternative)
    "*.sql",  # SQL (if desired)
]

# Extended list of directories and patterns to exclude.
_EXCLUDED_PATTERNS: list[str] = [
    "venv",  # Python virtual environments
    ".venv",
    "node_modules",  # Node.js dependencies
    ".git",  # Git repositories
    "__pycache__",  # Python cache directories
    "build",  # Build directories
    "dist",  # Distribution directories
    "target",  # Build output directories (Java, Rust, etc.)
    "bin",  # Binary directories
    "obj",  # Object files directories
    ".idea",  # JetBrains IDEs
    ".vscode",  # VS Code settings
    ".eggs",  # Python eggs
]


def list_files(
    directory: str = ".",
    included_patterns: list[str] = _INCLUDED_PATTERNS,
    excluded_patterns: list[str] = _EXCLUDED_PATTERNS,
) -> list[str]:
    """List all files in a directory that match any of the included glob patterns
    and do not reside in any directory matching an excluded pattern.
    Patterns are evaluated using glob-style matching.
    included_patterns and excluded_patterns already has sane default values.
    """
    all_files: list[str] = []
    for root, dirs, files in os.walk(directory):
        for filename in files:
            if any(fnmatch.fnmatch(filename, pat) for pat in included_patterns):
                full_path = os.path.join(root, filename)
                if _should_exclude(full_path, excluded_patterns):
                    continue
                all_files.append(full_path)
    return all_files


def _should_exclude(full_path: str, excluded_patterns: list[str]) -> bool:
    """
    Return True if the file at full_path should be excluded based on
    the list of excluded_patterns. Patterns that include a path separator
    are applied to the full normalized path; otherwise they are matched
    against each individual component of the path.
    """
    norm_path = os.path.normpath(full_path)
    path_parts = norm_path.split(os.sep)
    for pat in excluded_patterns:
        # If the pattern seems intended for full path matching (contains a separator)
        if os.sep in pat or "/" in pat:
            if fnmatch.fnmatch(norm_path, pat):
                return True
        else:
            # Otherwise check each part of the path
            if any(fnmatch.fnmatch(part, pat) for part in path_parts):
                return True
    return False


def read_text_file(file: str) -> str:
    """Read a text file and return a string containing the file content."""
    return read_file(os.path.abspath(file))


def write_text_file(file: str, content: str):
    """Write content to a text file"""
    return write_file(os.path.abspath(file), content)


def read_all_files(
    directory: str = ".",
    included_patterns: list[str] = _INCLUDED_PATTERNS,
    excluded_patterns: list[str] = _EXCLUDED_PATTERNS,
) -> list[str]:
    """Read all files in a directory that match any of the included glob patterns
    and do not match any of the excluded glob patterns.
    Patterns are evaluated using glob-style matching.
    included_patterns and excluded_patterns already has sane default values.
    """
    files = list_files(directory, included_patterns, excluded_patterns)
    for index, file in enumerate(files):
        content = read_text_file(file)
        files[index] = f"# {file}\n```\n{content}\n```"
    return files



import shlex


def get_remote_cmd_script(
    cmd_script: str,
    host: str = "",
    port: int | str = 22,
    user: str = "",
    password: str = "",
    use_password: bool = False,
    ssh_key: str = "",
) -> str:
    quoted_script = shlex.quote(cmd_script)
    if ssh_key != "" and use_password:
        return f'sshpass -p "{password}" ssh -t -p "{port}" -i "{ssh_key}" "{user}@{host}" {quoted_script}'  # noqa
    if ssh_key != "":
        return f'ssh -t -p "{port}" -i "{ssh_key}" "{user}@{host}" {quoted_script}'
    if use_password:
        return f'sshpass -p "{password}" ssh -t -p "{port}" "{user}@{host}" {quoted_script}'  # noqa
    return f'ssh -t -p "{port}" "{user}@{host}" {quoted_script}'



from zrb.builtin.group import base64_group
from zrb.context.any_context import AnyContext
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task


@make_task(
    name="encode-base64",
    input=StrInput(name="text", description="Text", prompt="Text to encode"),
    description="🔐 Encode text to base64",
    group=base64_group,
    alias="encode",
)
def encode_base64(ctx: AnyContext) -> str:
    import base64

    result = base64.b64encode(ctx.input.text.encode()).decode()
    ctx.print(result)
    return result


@make_task(
    name="decode-base64",
    input=StrInput(name="text", description="Text", prompt="Text to encode"),
    description="🔓 Decode base64 text",
    group=base64_group,
    alias="decode",
)
def decode_base64(ctx: AnyContext) -> str:
    import base64

    result = base64.b64decode(ctx.input.text.encode()).decode()
    ctx.print(result)
    return result



import datetime
from contextlib import asynccontextmanager
from typing import Any, Callable, Generic, Type, TypeVar

import ulid
from my_app_name.common.error import InvalidValueError, NotFoundError
from my_app_name.common.parser_factory import (
    parse_filter_param as default_parse_filter_param,
)
from my_app_name.common.parser_factory import (
    parse_sort_param as default_parse_sort_param,
)
from sqlalchemy import Engine, delete, func, insert, select, update
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession
from sqlalchemy.sql import ClauseElement, ColumnElement, Select
from sqlmodel import Session, SQLModel

DBModel = TypeVar("DBModel", bound=SQLModel)
ResponseModel = TypeVar("ResponseModel", bound=SQLModel)
CreateModel = TypeVar("CreateModel", bound=SQLModel)
UpdateModel = TypeVar("UpdateModel", bound=SQLModel)


class BaseDBRepository(Generic[DBModel, ResponseModel, CreateModel, UpdateModel]):
    db_model: Type[DBModel]
    response_model: Type[ResponseModel]
    create_model: Type[CreateModel]
    update_model: Type[UpdateModel]
    entity_name: str = "entity"
    column_preprocessors: dict[str, Callable[[Any], Any]] = {}

    def __init__(
        self,
        engine: Engine | AsyncEngine,
        filter_param_parser: (
            Callable[[SQLModel, str], list[ClauseElement]] | None
        ) = None,
        sort_param_parser: Callable[[SQLModel, str], list[ColumnElement]] | None = None,
    ):
        self.engine = engine
        self._is_async = isinstance(engine, AsyncEngine)
        self._parse_filter_param = (
            filter_param_parser if filter_param_parser else default_parse_filter_param
        )
        self._parse_sort_param = (
            sort_param_parser if sort_param_parser else default_parse_sort_param
        )

    @property
    def is_async(self) -> bool:
        return self._is_async

    def _select(self) -> Select:
        """
        This method is used to contruct select statement for get, get_by_id and get_by_ids.
        To parse the result of the statement, make sure you override _rows_to_response as well.
        """
        return select(self.db_model)

    def _rows_to_responses(self, rows: list[tuple[Any, ...]]) -> list[ResponseModel]:
        """
        This method is used to parse the result of select statement generated by _select.
        """
        return [self.response_model.model_validate(row[0]) for row in rows]

    async def _select_to_response(
        self, query_modifier: Callable[[Select], Any]
    ) -> list[ResponseModel]:
        statement = query_modifier(self._select())
        async with self._session_scope() as session:
            result = await self._execute_statement(session, statement)
            return self._rows_to_responses(result.all())

    def _ensure_one(self, responses: list[ResponseModel]) -> ResponseModel:
        if not responses:
            raise NotFoundError(f"{self.entity_name} not found")
        if len(responses) > 1:
            raise InvalidValueError(f"Duplicate {self.entity_name}")
        return responses[0]

    def _model_to_data_dict(
        self, data: SQLModel, **additional_data: Any
    ) -> dict[str, Any]:
        """
        This method transform SQLModel into dictionary for insert/update operation.
        """
        data_dict = data.model_dump(exclude_unset=True)
        data_dict.update(additional_data)
        for key, preprocessor in self.column_preprocessors.items():
            if key not in data_dict:
                continue
            if not hasattr(self.db_model, key):
                raise InvalidValueError(f"Invalid {self.entity_name} property: {key}")
            data_dict[key] = preprocessor(data_dict[key])
        return data_dict

    @asynccontextmanager
    async def _session_scope(self):
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                async with session.begin():
                    yield session
        else:
            with Session(self.engine) as session:
                with session.begin():
                    yield session

    async def _commit(self, session: Session | AsyncSession):
        if self.is_async:
            await session.commit()
        else:
            session.commit()

    async def _execute_statement(self, session, statement: Any) -> Any:
        if self.is_async:
            return await session.execute(statement)
        else:
            return session.execute(statement)

    async def get_by_id(self, id: str) -> ResponseModel:
        rows = await self._select_to_response(lambda q: q.where(self.db_model.id == id))
        return self._ensure_one(rows)

    async def get_by_ids(self, id_list: list[str]) -> list[ResponseModel]:
        rows = await self._select_to_response(
            lambda q: q.where(self.db_model.id.in_(id_list))
        )
        # raise error if any id not in id_list
        existing_id_list = [row.id for row in rows]
        inexist_id_list = [id for id in id_list if id not in existing_id_list]
        if len(inexist_id_list) > 0:
            raise NotFoundError(
                f"{self.entity_name} not found, inexist ids: {', '.join(inexist_id_list)}"
            )
        # sort rows
        row_dict = {row.id: row for row in rows}
        return [row_dict[id] for id in id_list]

    async def count(self, filter: str | None = None) -> int:
        count_statement = select(func.count(1)).select_from(self.db_model)
        if filter:
            filter_param = self._parse_filter_param(self.db_model, filter)
            count_statement = count_statement.where(*filter_param)
        async with self._session_scope() as session:
            result = await self._execute_statement(session, count_statement)
            return result.scalar_one()

    async def get(
        self,
        page: int = 1,
        page_size: int = 10,
        filter: str | None = None,
        sort: str | None = None,
    ) -> list[ResponseModel]:
        return await self._select_to_response(
            self._get_pagination_query_modifier(
                page=page, page_size=page_size, filter=filter, sort=sort
            )
        )

    def _get_pagination_query_modifier(
        self,
        page: int = 1,
        page_size: int = 10,
        filter: str | None = None,
        sort: str | None = None,
    ) -> Callable[[Select], Any]:
        def pagination_query_modifier(statement: Select) -> Any:
            offset = (page - 1) * page_size
            statement = statement.offset(offset).limit(page_size)
            if filter:
                filter_param = self._parse_filter_param(self.db_model, filter)
                statement = statement.where(*filter_param)
            if sort:
                sort_param = self._parse_sort_param(self.db_model, sort)
                statement = statement.order_by(*sort_param)
            return statement

        return pagination_query_modifier

    async def create(self, data: CreateModel) -> DBModel:
        now = datetime.datetime.now(datetime.timezone.utc)
        data_dict = self._model_to_data_dict(data, created_at=now, id=ulid.new().str)
        async with self._session_scope() as session:
            await self._execute_statement(
                session, insert(self.db_model).values(**data_dict)
            )
            statement = select(self.db_model).where(self.db_model.id == data_dict["id"])
            result = await self._execute_statement(session, statement)
            created_entity = result.scalar_one_or_none()
            if created_entity is None:
                raise NotFoundError(f"{self.entity_name} not found after creation")
            return self.db_model(**created_entity.model_dump())

    async def create_bulk(self, data_list: list[CreateModel]) -> list[DBModel]:
        now = datetime.datetime.now(datetime.timezone.utc)
        data_dict_list = [
            self._model_to_data_dict(data, created_at=now, id=ulid.new().str)
            for data in data_list
        ]
        id_list = [data_dict["id"] for data_dict in data_dict_list]
        async with self._session_scope() as session:
            await self._execute_statement(
                session, insert(self.db_model).values(data_dict_list)
            )
            statement = select(self.db_model).where(self.db_model.id.in_(id_list))
            result = await self._execute_statement(session, statement)
            row_dict = {
                entity.id: self.db_model(**entity.model_dump())
                for entity in result.scalars().all()
            }
            return [row_dict[id] for id in id_list]

    async def delete(self, id: str) -> DBModel:
        async with self._session_scope() as session:
            statement = select(self.db_model).where(self.db_model.id == id)
            result = await self._execute_statement(session, statement)
            entity = result.scalar_one_or_none()
            if not entity:
                raise NotFoundError(f"{self.entity_name} not found")
            await self._execute_statement(
                session, delete(self.db_model).where(self.db_model.id == id)
            )
            return self.db_model(**entity.model_dump())

    async def delete_bulk(self, id_list: list[str]) -> list[DBModel]:
        async with self._session_scope() as session:
            statement = select(self.db_model).where(self.db_model.id.in_(id_list))
            result = await self._execute_statement(session, statement)
            entities = result.scalars().all()
            await self._execute_statement(
                session, delete(self.db_model).where(self.db_model.id.in_(id_list))
            )
            row_dict = {
                entity.id: self.db_model(**entity.model_dump()) for entity in entities
            }
            return [row_dict[id] for id in id_list]

    async def update(self, id: str, data: UpdateModel) -> DBModel:
        now = datetime.datetime.now(datetime.timezone.utc)
        update_data = self._model_to_data_dict(data, updated_at=now)
        update_data = {k: v for k, v in update_data.items() if v is not None}
        async with self._session_scope() as session:
            statement = (
                update(self.db_model)
                .where(self.db_model.id == id)
                .values(**update_data)
            )
            await self._execute_statement(session, statement)
            result = await self._execute_statement(
                session, select(self.db_model).where(self.db_model.id == id)
            )
            updated_instance = result.scalar_one_or_none()
            if not updated_instance:
                raise NotFoundError(f"{self.entity_name} not found")
            return self.db_model(**updated_instance.model_dump())

    async def update_bulk(self, id_list: list[str], data: UpdateModel) -> list[DBModel]:
        now = datetime.datetime.now(datetime.timezone.utc)
        update_data = self._model_to_data_dict(data, updated_at=now)
        update_data = {k: v for k, v in update_data.items() if v is not None}
        if not update_data:
            raise InvalidValueError("No valid update data provided")
        async with self._session_scope() as session:
            statement = (
                update(self.db_model)
                .where(self.db_model.id.in_(id_list))
                .values(**update_data)
            )
            await self._execute_statement(session, statement)
            result = await self._execute_statement(
                session, select(self.db_model).where(self.db_model.id.in_(id_list))
            )
            row_dict = {
                entity.id: self.db_model(**entity.model_dump())
                for entity in result.scalars().all()
            }
            return [row_dict[id] for id in id_list]



import asyncio
import os
import re
import sys
from collections.abc import Callable

import psutil

from zrb.cmd.cmd_result import CmdResult


def check_unrecommended_commands(cmd_script: str) -> dict[str, str]:
    banned_commands = {
        "<(": "Process substitution isn't POSIX compliant and causes trouble",
        "column": "Command isn't included in Ubuntu packages and is not POSIX compliant",
        "echo": "echo isn't consistent across OS; use printf instead",
        "eval": "Avoid eval as it can accidentally execute arbitrary strings",
        "realpath": "Not available by default on OSX",
        "source": "Not POSIX compliant; use '.' instead",
        " test": "Use '[' instead for consistency",
        "which": "Command in not POSIX compliant, use command -v",
    }
    banned_commands_regex = {
        r"grep.* -y": "grep -y does not work on Alpine; use grep -i",
        r"grep.* -P": "grep -P is not valid on OSX",
        r"grep[^|]+--\w{2,}": "grep long commands do not work on Alpine",
        r'readlink.+-.*f.+["$]': "readlink -f behaves differently on OSX",
        r"sort.*-V": "sort -V is not supported everywhere",
        r"sort.*--sort-versions": "sort --sort-version is not supported everywhere",
        r"\bls ": "Avoid using ls; use shell globs or find instead",
    }
    violations = {}
    # Check banned commands
    for cmd, reason in banned_commands.items():
        if cmd in cmd_script:
            violations[cmd] = reason
    # Check banned regex patterns
    for pattern, reason in banned_commands_regex.items():
        if re.search(pattern, cmd_script):
            violations[pattern] = reason
    return violations


async def run_command(
    cmd: list[str],
    cwd: str | None = None,
    env_map: dict[str, str] | None = None,
    print_method: Callable[..., None] | None = None,
    register_pid_method: Callable[[int], None] | None = None,
    max_output_line: int = 1000,
    max_error_line: int = 1000,
) -> tuple[CmdResult, int]:
    async def __read_stream(
        stream, print_method: Callable[..., None], max_lines: int
    ) -> str:
        lines = []
        while True:
            line = await stream.readline()
            if not line:
                break
            line = line.decode("utf-8").rstrip()
            lines.append(line)
            if len(lines) > max_lines:
                lines.pop(0)  # Keep only the last max_lines
            print_method(line)
        return "\n".join(lines)

    actual_print_method = print_method if print_method is not None else print
    cmd_process = None
    if cwd is None:
        cwd = os.getcwd()
    if env_map is None:
        env_map = os.environ
    cmd_process = await asyncio.create_subprocess_exec(
        *cmd,
        cwd=cwd,
        stdin=sys.stdin if sys.stdin.isatty() else None,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        env=env_map,
        bufsize=0,
    )
    if register_pid_method is not None:
        register_pid_method(cmd_process.pid)
    stdout_task = asyncio.create_task(
        __read_stream(cmd_process.stdout, actual_print_method, max_output_line)
    )
    stderr_task = asyncio.create_task(
        __read_stream(cmd_process.stderr, actual_print_method, max_error_line)
    )
    # Wait for process to complete and gather stdout/stderr
    return_code = await cmd_process.wait()
    stdout = await stdout_task
    stderr = await stderr_task
    return CmdResult(stdout, stderr), return_code


def kill_pid(pid: int, print_method: Callable[..., None] | None = None):
    actual_print_method = print_method if print_method is not None else print
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    for child in children:
        actual_print_method(f"Killing child process {child.pid}")
        child.terminate()
    actual_print_method(f"Killing process {pid}")
    parent.terminate()



import fnmatch
import hashlib
import json
import os
import sys
from collections.abc import Callable

import ulid

from zrb.config import (
    RAG_CHUNK_SIZE,
    RAG_EMBEDDING_API_KEY,
    RAG_EMBEDDING_BASE_URL,
    RAG_EMBEDDING_MODEL,
    RAG_MAX_RESULT_COUNT,
    RAG_OVERLAP,
)
from zrb.util.cli.style import stylize_error, stylize_faint
from zrb.util.file import read_file


class RAGFileReader:
    def __init__(self, glob_pattern: str, read: Callable[[str], str]):
        self.glob_pattern = glob_pattern
        self.read = read

    def is_match(self, file_name: str):
        if os.sep not in self.glob_pattern and (
            os.altsep is None or os.altsep not in self.glob_pattern
        ):
            # Pattern like "*.txt" – match only the basename.
            return fnmatch.fnmatch(os.path.basename(file_name), self.glob_pattern)
        return fnmatch.fnmatch(file_name, self.glob_pattern)


def create_rag_from_directory(
    tool_name: str,
    tool_description: str,
    document_dir_path: str = "./documents",
    vector_db_path: str = "./chroma",
    vector_db_collection: str = "documents",
    chunk_size: int = RAG_CHUNK_SIZE,
    overlap: int = RAG_OVERLAP,
    max_result_count: int = RAG_MAX_RESULT_COUNT,
    file_reader: list[RAGFileReader] = [],
    openai_api_key: str = RAG_EMBEDDING_API_KEY,
    openai_base_url: str = RAG_EMBEDDING_BASE_URL,
    openai_embedding_model: str = RAG_EMBEDDING_MODEL,
):
    async def retrieve(query: str) -> str:
        from chromadb import PersistentClient
        from chromadb.config import Settings
        from openai import OpenAI

        # Initialize OpenAI client with custom URL if provided
        client_args = {}
        if openai_api_key:
            client_args["api_key"] = openai_api_key
        if openai_base_url:
            client_args["base_url"] = openai_base_url
        # Initialize OpenAI client for embeddings
        openai_client = OpenAI(**client_args)
        # Initialize ChromaDB client
        chroma_client = PersistentClient(
            path=vector_db_path, settings=Settings(allow_reset=True)
        )
        collection = chroma_client.get_or_create_collection(vector_db_collection)
        # Track file changes using a hash-based approach
        hash_file_path = os.path.join(vector_db_path, "file_hashes.json")
        previous_hashes = _load_hashes(hash_file_path)
        current_hashes = {}
        # Get updated_files
        updated_files = []
        for root, _, files in os.walk(document_dir_path):
            for file in files:
                file_path = os.path.join(root, file)
                file_hash = _compute_file_hash(file_path)
                relative_path = os.path.relpath(file_path, document_dir_path)
                current_hashes[relative_path] = file_hash
                if previous_hashes.get(relative_path) != file_hash:
                    updated_files.append(file_path)
        # Upsert updated_files to vector db
        if updated_files:
            print(
                stylize_faint(f"Updating {len(updated_files)} changed files"),
                file=sys.stderr,
            )
            for file_path in updated_files:
                try:
                    relative_path = os.path.relpath(file_path, document_dir_path)
                    collection.delete(where={"file_path": relative_path})
                    content = _read_txt_content(file_path, file_reader)
                    file_id = ulid.new().str
                    for i in range(0, len(content), chunk_size - overlap):
                        chunk = content[i : i + chunk_size]
                        if chunk:
                            chunk_id = ulid.new().str
                            print(
                                stylize_faint(
                                    f"Vectorizing {relative_path} chunk {chunk_id}"
                                ),
                                file=sys.stderr,
                            )
                            # Get embeddings using OpenAI
                            embedding_response = openai_client.embeddings.create(
                                input=chunk, model=openai_embedding_model
                            )
                            vector = embedding_response.data[0].embedding
                            collection.upsert(
                                ids=[chunk_id],
                                embeddings=[vector],
                                documents=[chunk],
                                metadatas={
                                    "file_path": relative_path,
                                    "file_id": file_id,
                                },
                            )
                except Exception as e:
                    print(
                        stylize_error(f"Error processing {file_path}: {e}"),
                        file=sys.stderr,
                    )
            _save_hashes(hash_file_path, current_hashes)
        else:
            print(
                stylize_faint("No changes detected. Skipping database update."),
                file=sys.stderr,
            )
        # Vectorize query and get related document chunks
        print(stylize_faint("Vectorizing query"), file=sys.stderr)
        # Get embeddings using OpenAI
        embedding_response = openai_client.embeddings.create(
            input=query, model=openai_embedding_model
        )
        query_vector = embedding_response.data[0].embedding
        print(stylize_faint("Searching documents"), file=sys.stderr)
        results = collection.query(
            query_embeddings=query_vector,
            n_results=max_result_count,
        )
        return json.dumps(results)

    retrieve.__name__ = tool_name
    retrieve.__doc__ = tool_description
    return retrieve


def _compute_file_hash(file_path: str) -> str:
    hash_md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def _load_hashes(file_path: str) -> dict:
    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            return json.load(f)
    return {}


def _save_hashes(file_path: str, hashes: dict):
    with open(file_path, "w") as f:
        json.dump(hashes, f)


def _read_txt_content(file_path: str, file_reader: list[RAGFileReader]):
    for reader in file_reader:
        if reader.is_match(file_path):
            return reader.read(file_path)
    if file_path.lower().endswith(".pdf"):
        return _read_pdf(file_path)
    return read_file(file_path)


def _read_pdf(file_path: str) -> str:
    import pdfplumber

    with pdfplumber.open(file_path) as pdf:
        return "\n".join(
            page.extract_text() for page in pdf.pages if page.extract_text()
        )



from zrb.builtin.group import md5_group
from zrb.context.any_context import AnyContext
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task


@make_task(
    name="hash-md5",
    input=StrInput(name="text", description="Text", prompt="Text to encode"),
    description="🧩 Hash text to md5",
    group=md5_group,
    alias="hash",
)
def hash_md5(ctx: AnyContext) -> str:
    import hashlib

    result = hashlib.md5(ctx.input.text.encode()).hexdigest()
    ctx.print(result)
    return result


@make_task(
    name="sum-md5",
    input=StrInput(name="file", description="File name", prompt="File name"),
    description="➕ Get md5 checksum of a file",
    group=md5_group,
    alias="sum",
)
def sum_md5(ctx: AnyContext) -> str:
    import hashlib

    with open(ctx.input.file, mode="rb") as file:
        content = file.read()
    result = hashlib.md5(content).hexdigest()
    ctx.print(result)
    return result



from zrb.group.any_group import AnyGroup
from zrb.util.group import get_non_empty_subgroups, get_subtasks


class SubCommand:
    def __init__(self, paths: list[str] = [], nexts: list[str] = []):
        self.paths = paths
        self.nexts = nexts

    def __repr__(self):
        return f"<{self.__class__.__name__} paths={self.paths} nexts={self.nexts}>"


def get_group_subcommands(
    group: AnyGroup, previous_path: str = [], subcommands: list[SubCommand] = []
) -> list[SubCommand]:
    nexts = []
    for task_alias in get_subtasks(group):
        nexts.append(task_alias)
    for subgroup_alias, subgroup in get_non_empty_subgroups(group).items():
        nexts.append(subgroup_alias)
        # Recursively add subgroup
        get_group_subcommands(
            group=subgroup,
            previous_path=previous_path + [group.name],
            subcommands=subcommands,
        )
    if len(nexts) > 0:
        subcommands.append(SubCommand(paths=previous_path + [group.name], nexts=nexts))
    return subcommands



import datetime
import json
import os
from typing import Any

from zrb.builtin.group import todo_group
from zrb.config import TODO_DIR, TODO_RETENTION, TODO_VISUAL_FILTER
from zrb.context.any_context import AnyContext
from zrb.input.str_input import StrInput
from zrb.input.text_input import TextInput
from zrb.task.make_task import make_task
from zrb.util.file import read_file, write_file
from zrb.util.todo import (
    TodoTaskModel,
    add_duration,
    cascade_todo_task,
    get_visual_todo_card,
    get_visual_todo_list,
    line_to_todo_task,
    load_todo_list,
    parse_duration,
    save_todo_list,
    select_todo_task,
    todo_task_to_line,
)


@make_task(
    name="add-todo",
    input=[
        StrInput(
            name="description",
            description="Task description",
            prompt="Task description",
        ),
        StrInput(
            name="priority",
            description="Task priority",
            prompt="Task priority",
            default="E",
        ),
        StrInput(
            name="project",
            description="Task project",
            prompt="Task project (space separated)",
            allow_empty=True,
        ),
        StrInput(
            name="context",
            description="Task context",
            prompt="Task context (space separated)",
            allow_empty=True,
        ),
    ],
    description="➕ Add todo",
    group=todo_group,
    alias="add",
)
def add_todo(ctx: AnyContext):
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    todo_list: list[TodoTaskModel] = []
    if os.path.isfile(todo_file_path):
        todo_list = load_todo_list(todo_file_path)
    else:
        os.makedirs(TODO_DIR, exist_ok=True)
    todo_list.append(
        cascade_todo_task(
            TodoTaskModel(
                priority=ctx.input.priority.upper(),
                description=ctx.input.description,
                contexts=[
                    context.strip()
                    for context in ctx.input.context.split(" ")
                    if context.strip() != ""
                ],
                projects=[
                    project.strip()
                    for project in ctx.input.project.split(" ")
                    if project.strip() != ""
                ],
            )
        )
    )
    save_todo_list(todo_file_path, todo_list)
    return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)


@make_task(name="list-todo", description="📋 List todo", group=todo_group, alias="list")
def list_todo(ctx: AnyContext):
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    todo_list: list[TodoTaskModel] = []
    if os.path.isfile(todo_file_path):
        todo_list = load_todo_list(todo_file_path)
    return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)


@make_task(
    name="show-todo",
    input=StrInput(name="keyword", prompt="Task keyword", description="Task Keyword"),
    description="🔍 Show todo",
    group=todo_group,
    alias="show",
)
def show_todo(ctx: AnyContext):
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    todo_list: list[TodoTaskModel] = []
    todo_list: list[TodoTaskModel] = []
    if os.path.isfile(todo_file_path):
        todo_list = load_todo_list(todo_file_path)
    # Get todo task
    todo_task = select_todo_task(todo_list, ctx.input.keyword)
    if todo_task is None:
        ctx.log_error("Task not found")
        return
    if todo_task.completed:
        ctx.log_error("Task already completed")
        return
    # Update todo task
    todo_task = cascade_todo_task(todo_task)
    task_id = todo_task.keyval.get("id", "")
    log_work_path = os.path.join(TODO_DIR, "log-work", f"{task_id}.json")
    log_work_list = []
    if os.path.isfile(log_work_path):
        log_work_list = json.loads(read_file(log_work_path))
    return get_visual_todo_card(todo_task, log_work_list)


@make_task(
    name="complete-todo",
    input=StrInput(name="keyword", prompt="Task keyword", description="Task Keyword"),
    description="✅ Complete todo",
    group=todo_group,
    alias="complete",
)
def complete_todo(ctx: AnyContext):
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    todo_list: list[TodoTaskModel] = []
    if os.path.isfile(todo_file_path):
        todo_list = load_todo_list(todo_file_path)
    # Get todo task
    todo_task = select_todo_task(todo_list, ctx.input.keyword)
    if todo_task is None:
        ctx.log_error("Task not found")
        return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)
    if todo_task.completed:
        ctx.log_error("Task already completed")
        return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)
    # Update todo task
    todo_task = cascade_todo_task(todo_task)
    if todo_task.creation_date is not None:
        todo_task.completion_date = datetime.date.today()
    todo_task.completed = True
    # Save todo list
    save_todo_list(todo_file_path, todo_list)
    return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)


@make_task(
    name="archive-todo",
    description="📚 Archive todo",
    group=todo_group,
    alias="archive",
)
def archive_todo(ctx: AnyContext):
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    todo_list: list[TodoTaskModel] = []
    if os.path.isfile(todo_file_path):
        todo_list = load_todo_list(todo_file_path)
    retention_duration = datetime.timedelta(seconds=parse_duration(TODO_RETENTION))
    threshold_date = datetime.date.today() - retention_duration
    new_archived_todo_list = [
        todo_task
        for todo_task in todo_list
        if todo_task.completed
        and todo_task.completion_date is not None
        and todo_task.completion_date < threshold_date
    ]
    working_todo_list = [
        todo_task for todo_task in todo_list if todo_task not in new_archived_todo_list
    ]
    if len(new_archived_todo_list) == 0:
        ctx.print("No completed task to archive")
        return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)
    archive_file_path = os.path.join(TODO_DIR, "archive.txt")
    if not os.path.isdir(TODO_DIR):
        os.make_dirs(TODO_DIR, exist_ok=True)
    # Get archived todo list
    archived_todo_list = []
    if os.path.isfile(archive_file_path):
        archived_todo_list = load_todo_list(archive_file_path)
    archived_todo_list += new_archived_todo_list
    # Save the new todo list and add the archived ones
    save_todo_list(archive_file_path, archived_todo_list)
    save_todo_list(todo_file_path, working_todo_list)
    return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)


@make_task(
    name="log-todo",
    input=[
        StrInput(name="keyword", prompt="Task keyword", description="Task Keyword"),
        StrInput(
            name="log",
            prompt="Working log",
            description="Working log",
        ),
        StrInput(
            name="duration",
            prompt="Working duration",
            description="Working duration",
            default="30m",
        ),
        StrInput(
            name="stop",
            prompt="Working stop time (%Y-%m-%d %H:%M:%S)",
            description="Working stop time",
            default=lambda _: _get_default_stop_work_time_str(),
        ),
    ],
    description="🕒 Log work todo",
    group=todo_group,
    alias="log",
)
def log_todo(ctx: AnyContext):
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    todo_list: list[TodoTaskModel] = []
    if os.path.isfile(todo_file_path):
        todo_list = load_todo_list(todo_file_path)
    # Get todo task
    todo_task = select_todo_task(todo_list, ctx.input.keyword)
    if todo_task is None:
        ctx.log_error("Task not found")
        return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)
    # Update todo task
    todo_task = cascade_todo_task(todo_task)
    current_duration_str = todo_task.keyval.get("duration", "0")
    todo_task.keyval["duration"] = add_duration(
        current_duration_str, ctx.input.duration
    )
    # Save todo list
    save_todo_list(todo_file_path, todo_list)
    # Add log work
    log_work_dir = os.path.join(TODO_DIR, "log-work")
    os.makedirs(log_work_dir, exist_ok=True)
    log_work_file_path = os.path.join(
        log_work_dir, f"{todo_task.keyval.get('id')}.json"
    )
    if os.path.isfile(log_work_file_path):
        log_work_json = read_file(log_work_file_path)
    else:
        log_work_json = "[]"
    log_work: list[dict[str, Any]] = json.loads(log_work_json)
    start_work_time_str = _get_start_work_time_str(ctx.input.stop, ctx.input.duration)
    log_work.append(
        {
            "log": ctx.input.log,
            "duration": ctx.input.duration,
            "start": start_work_time_str,
        }
    )
    # save todo with log work
    write_file(log_work_file_path, json.dumps(log_work, indent=2))
    # get log work list
    task_id = todo_task.keyval.get("id", "")
    log_work_path = os.path.join(TODO_DIR, "log-work", f"{task_id}.json")
    log_work_list = []
    if os.path.isfile(log_work_path):
        log_work_list = json.loads(read_file(log_work_path))
    return "\n".join(
        [
            get_visual_todo_list(todo_list, TODO_VISUAL_FILTER),
            "",
            get_visual_todo_card(todo_task, log_work_list),
        ]
    )


def _get_start_work_time_str(stop_work_time_str: str, work_duration_str: str) -> str:
    work_duration = parse_duration(work_duration_str)
    stop_work_time = datetime.datetime.strptime(stop_work_time_str, "%Y-%m-%d %H:%M:%S")
    start_work_time = stop_work_time - datetime.timedelta(seconds=work_duration)
    return start_work_time.strftime("%Y-%m-%d %H:%M:%S")


def _get_default_stop_work_time_str() -> str:
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")


@make_task(
    name="edit-todo",
    input=[
        TextInput(
            name="text",
            description="Todo.txt content",
            prompt="Todo.txt content (will override existing)",
            default=lambda _: _get_todo_txt_content(),
            allow_positional_parsing=False,
        ),
    ],
    description="📝 Edit todo",
    group=todo_group,
    alias="edit",
)
def edit_todo(ctx: AnyContext):
    todo_list = [
        cascade_todo_task(line_to_todo_task(line))
        for line in ctx.input.text.split("\n")
        if line.strip() != ""
    ]
    new_content = "\n".join(todo_task_to_line(todo_task) for todo_task in todo_list)
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    write_file(todo_file_path, new_content)
    todo_list = load_todo_list(todo_file_path)
    return get_visual_todo_list(todo_list, TODO_VISUAL_FILTER)


def _get_todo_txt_content() -> str:
    todo_file_path = os.path.join(TODO_DIR, "todo.txt")
    if not os.path.isfile(todo_file_path):
        return ""
    return read_file(todo_file_path)



from zrb.attr.type import BoolAttr
from zrb.context.any_shared_context import AnySharedContext
from zrb.input.base_input import BaseInput
from zrb.util.attr import get_bool_attr
from zrb.util.string.conversion import to_boolean


class BoolInput(BaseInput):
    def __init__(
        self,
        name: str,
        description: str | None = None,
        prompt: str | None = None,
        default: BoolAttr = False,
        auto_render: bool = True,
        allow_empty: bool = False,
        allow_positional_parsing: bool = True,
        always_prompt: bool = True,
    ):
        super().__init__(
            name=name,
            description=description,
            prompt=prompt,
            default=default,
            auto_render=auto_render,
            allow_empty=allow_empty,
            allow_positional_parsing=allow_positional_parsing,
            always_prompt=always_prompt,
        )

    def to_html(self, shared_ctx: AnySharedContext) -> str:
        name = self.name
        description = self.description
        default = to_boolean(self.get_default_str(shared_ctx))
        selected_true = "selected" if default else ""
        selected_false = "selected" if not default else ""
        return "\n".join(
            [
                f'<select name="{name}" placeholder="{description}">',
                f'<option value="true" {selected_true}>true</option>',
                f'<option value="false" {selected_false}>false</option>',
                "</select>",
            ]
        )

    def get_default_str(self, shared_ctx: AnySharedContext) -> str:
        default_value = get_bool_attr(
            shared_ctx, self._default_value, auto_render=self._auto_render
        )
        return f"{default_value}"

    def _parse_str_value(self, str_value: str) -> bool:
        return to_boolean(str_value)



from zrb.attr.type import (
    AnyAttr,
    BoolAttr,
    FloatAttr,
    IntAttr,
    StrAttr,
    StrDictAttr,
    fstring,
)
from zrb.callback.any_callback import AnyCallback
from zrb.callback.callback import Callback
from zrb.cmd.cmd_result import CmdResult
from zrb.cmd.cmd_val import Cmd, CmdPath
from zrb.config import LOAD_BUILTIN
from zrb.content_transformer.any_content_transformer import AnyContentTransformer
from zrb.content_transformer.content_transformer import ContentTransformer
from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext
from zrb.context.context import Context
from zrb.context.shared_context import SharedContext
from zrb.env.any_env import AnyEnv
from zrb.env.env import Env
from zrb.env.env_file import EnvFile
from zrb.env.env_map import EnvMap
from zrb.group.any_group import AnyGroup
from zrb.group.group import Group
from zrb.input.any_input import AnyInput
from zrb.input.base_input import BaseInput
from zrb.input.bool_input import BoolInput
from zrb.input.float_input import FloatInput
from zrb.input.int_input import IntInput
from zrb.input.option_input import OptionInput
from zrb.input.password_input import PasswordInput
from zrb.input.str_input import StrInput
from zrb.input.text_input import TextInput
from zrb.llm_config import llm_config
from zrb.runner.cli import cli
from zrb.runner.web_config.config_factory import web_config
from zrb.runner.web_schema.user import User
from zrb.session.session import Session
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.task.base_trigger import BaseTrigger
from zrb.task.cmd_task import CmdTask
from zrb.task.http_check import HttpCheck
from zrb.task.llm_task import LLMTask
from zrb.task.make_task import make_task
from zrb.task.rsync_task import RsyncTask
from zrb.task.scaffolder import Scaffolder
from zrb.task.scheduler import Scheduler
from zrb.task.task import Task
from zrb.task.tcp_check import TcpCheck
from zrb.util.load import load_file, load_module
from zrb.xcom.xcom import Xcom

assert load_file
assert load_module
assert fstring
assert AnyAttr
assert BoolAttr
assert IntAttr
assert FloatAttr
assert StrAttr
assert StrDictAttr
assert AnyCallback
assert Callback
assert AnyEnv
assert Env
assert EnvFile
assert EnvMap
assert AnyInput
assert BaseInput
assert BoolInput
assert Cmd
assert CmdPath
assert CmdResult
assert CmdTask
assert HttpCheck
assert TcpCheck
assert FloatInput
assert IntInput
assert OptionInput
assert PasswordInput
assert StrInput
assert TextInput
assert AnyGroup
assert Group
assert AnyTask
assert BaseTask
assert BaseTrigger
assert RsyncTask
assert Task
assert LLMTask
assert Session
assert AnyContext
assert Context
assert AnySharedContext
assert SharedContext
assert make_task
assert AnyContentTransformer
assert ContentTransformer
assert Scaffolder
assert Scheduler
assert cli
assert llm_config
assert Xcom
assert web_config
assert User

if LOAD_BUILTIN:
    from zrb import builtin

    assert builtin
